/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpidi_ch3_impl.h"
#include "pmi.h"
#include "sock.h"

volatile unsigned int MPIDI_CH3I_progress_completions = 0;

enum conn_state
{
    CONN_STATE_UNCONNECTED,
    CONN_STATE_LISTENING,
    CONN_STATE_CONNECTING,
    CONN_STATE_OPEN_CSEND,
    CONN_STATE_OPEN_CRECV,
    CONN_STATE_OPEN_LRECV,
    CONN_STATE_OPEN_LSEND,
    CONN_STATE_CONNECTED,
    CONN_STATE_CLOSING,
    CONN_STATE_CLOSED,
    CONN_STATE_FAILED
};

typedef struct MPIDI_CH3I_Connection
{
    MPIDI_VC * vc;
    sock_t sock;
    enum conn_state state;
    MPID_Request * send_active;
    MPID_Request * recv_active;
    MPIDI_CH3_Pkt_t pkt;
} MPIDI_CH3I_Connection_t;

static sock_set_t sock_set;
static int listener_port = 0;
static MPIDI_CH3I_Connection_t * listener_conn = NULL;

static int shutting_down = FALSE;

static inline MPIDI_CH3I_Connection_t * connection_alloc(void);
static inline void connection_free(MPIDI_CH3I_Connection_t * conn);
static inline void connection_post_sendq_req(MPIDI_CH3I_Connection_t * conn);
static inline void connection_post_send_pkt(MPIDI_CH3I_Connection_t * conn);
static inline void connection_post_recv_pkt(MPIDI_CH3I_Connection_t * conn);
static void connection_send_fail(MPIDI_CH3I_Connection_t * conn, int sock_errno);
static void connection_recv_fail(MPIDI_CH3I_Connection_t * conn, int sock_errno);
static int sock_errno_to_mpi_errno(int sock_errno);

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_start
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Progress_start()
{
    /* MT - This function is empty for the single-threaded implementation */
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_START);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_START);
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_START);
}

/*
 * MPIDI_CH3I_Request_adjust_iov()
 *
 * Adjust the iovec in the request by the supplied number of bytes.  If the iovec has been consumed, return true; otherwise return
 * false.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Request_adjust_iov
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Request_adjust_iov(MPID_Request * req, MPIDI_msg_sz_t nb)
{
    int offset = req->mm.iov_offset;
    const int count = req->ch3.iov_count;
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_REQUEST_ADJUST_IOV);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_REQUEST_ADJUST_IOV);
    
    while (offset < count)
    {
	if (req->ch3.iov[offset].MPID_IOV_LEN <= (unsigned int)nb)
	{
	    nb -= req->ch3.iov[offset].MPID_IOV_LEN;
	    offset++;
	}
	else
	{
	    (char *) req->ch3.iov[offset].MPID_IOV_BUF += nb;
	    req->ch3.iov[offset].MPID_IOV_LEN -= nb;
	    req->mm.iov_offset = offset;
	    MPIDI_DBG_PRINTF((60, FCNAME, "adjust_iov returning FALSE\n"));
	    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_REQUEST_ADJUST_IOV);
	    return FALSE;
	}
    }
    
    req->mm.iov_offset = offset;

    MPIDI_DBG_PRINTF((60, FCNAME, "adjust_iov returning TRUE\n"));
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_REQUEST_ADJUST_IOV);
    return TRUE;
}

#define post_pkt_recv(vc) vc->mm.shm_reading_pkt = TRUE

#undef FUNCNAME
#define FUNCNAME handle_shm_read
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void handle_shm_read(MPIDI_VC *vc, int nb)
{
    MPID_Request * req;
    MPIDI_STATE_DECL(MPID_STATE_HANDLE_SHM_READ);

    MPIDI_FUNC_ENTER(MPID_STATE_HANDLE_SHM_READ);
    
    MPIDI_DBG_PRINTF((60, FCNAME, "entering"));

    req = vc->mm.recv_active;
    if (req == NULL)
    {
	MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_SHM_READ);
	return;
    }

    if (nb > 0)
    {
	if (MPIDI_CH3I_Request_adjust_iov(req, nb))
	{
	    /* Read operation complete */
	    MPIDI_CA_t ca = req->ch3.ca;
	    
	    vc->mm.recv_active = NULL;
	    
	    /*if (ca == MPIDI_CH3I_CA_HANDLE_PKT)
	    {
		MPIDI_CH3_Pkt_t * pkt = &req->mm.pkt;
		
		if (pkt->type < MPIDI_CH3_PKT_END_CH3)
		{
		    printf("I should never get here anymore.\n");fflush(stdout);
		    MPIDI_DBG_PRINTF((65, FCNAME, "received CH3 packet %d, calllng CH3U_Handle_recv_pkt()", pkt->type));
		    MPIDI_CH3U_Handle_recv_pkt(vc, pkt);
		    MPIDI_DBG_PRINTF((65, FCNAME, "CH3U_Handle_recv_pkt() returned"));
		    if (vc->mm.recv_active == NULL)
		    {
			MPIDI_DBG_PRINTF((65, FCNAME, "complete; posting new recv packet"));
			post_pkt_recv(vc);
			MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
			MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_SHM_READ);
			return;
		    }
		}
	    }
	    else */ if (ca == MPIDI_CH3_CA_COMPLETE)
	    {
		MPIDI_DBG_PRINTF((65, FCNAME, "received requested data, decrementing CC"));
		/* mark data transfer as complete adn decrment CC */
		req->ch3.iov_count = 0;
		MPIDI_CH3U_Request_complete(req);
		post_pkt_recv(vc);
		MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
		MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_SHM_READ);
		return;
	    }
	    else if (ca < MPIDI_CH3_CA_END_CH3)
	    {
		/* XXX - This code assumes that if another read is not posted by the device during the callback, then the
		   device is not expecting any more data for request.  As a result, the channels posts a read for another
		   packet */
		MPIDI_DBG_PRINTF((65, FCNAME, "finished receiving iovec, calling CH3U_Handle_recv_req()"));
		MPIDI_CH3U_Handle_recv_req(vc, req);
		if (vc->mm.recv_active == NULL)
		{
		    MPIDI_DBG_PRINTF((65, FCNAME, "request (assumed) complete, posting new recv packet"));
		    post_pkt_recv(vc);
		    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
		    MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_SHM_READ);
		    return;
		}
	    }
	    else
	    {
		assert(ca != MPIDI_CH3I_CA_HANDLE_PKT);
		assert(ca < MPIDI_CH3_CA_END_CH3);
	    }
	}
	else
	{
	    assert(req->mm.iov_offset < req->ch3.iov_count);
	    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
	    MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_SHM_READ);
	    return;
	}
    }
    else
    {
	MPIDI_DBG_PRINTF((65, FCNAME, "Read args were iov=%x, count=%d",
	    req->ch3.iov + req->mm.iov_offset, req->ch3.iov_count - req->mm.iov_offset));
    }
    
    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
    MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_SHM_READ);
}

#undef FUNCNAME
#define FUNCNAME handle_shm_written
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void handle_shm_written(MPIDI_VC * vc)
{
    int error;
    int nb;
    MPIDI_STATE_DECL(MPID_STATE_HANDLE_SHM_WRITTEN);

    MPIDI_FUNC_ENTER(MPID_STATE_HANDLE_SHM_WRITTEN);
    
    MPIDI_DBG_PRINTF((60, FCNAME, "entering"));
    while (vc->mm.send_active != NULL)
    {
	MPID_Request * req = vc->mm.send_active;

	/*
	if (req->mm.iov_offset >= req->ch3.iov_count)
	{
	    MPIDI_DBG_PRINTF((60, FCNAME, "iov_offset(%d) >= iov_count(%d)", req->mm.iov_offset, req->ch3.iov_count));
	}
	*/
	assert(req->mm.iov_offset < req->ch3.iov_count);
	/*MPIDI_DBG_PRINTF((60, FCNAME, "calling shm_writev"));*/
	error = MPIDI_CH3I_SHM_writev(vc, req->ch3.iov + req->mm.iov_offset, req->ch3.iov_count - req->mm.iov_offset, &nb);
	MPIDI_DBG_PRINTF((60, FCNAME, "shm_writev returned %d", nb));

	if (nb > 0)
	{
	    if (MPIDI_CH3I_Request_adjust_iov(req, nb))
	    {
		/* Write operation complete */
		MPIDI_CA_t ca = req->ch3.ca;
			
		vc->mm.send_active = NULL;
		
		if (ca == MPIDI_CH3_CA_COMPLETE)
		{
		    MPIDI_DBG_PRINTF((65, FCNAME, "sent requested data, decrementing CC"));
		    MPIDI_CH3I_SendQ_dequeue(vc);
		    vc->mm.send_active = MPIDI_CH3I_SendQ_head(vc);
		    /* mark data transfer as complete and decrment CC */
		    req->ch3.iov_count = 0;
		    MPIDI_CH3U_Request_complete(req);
		}
		else if (ca == MPIDI_CH3I_CA_HANDLE_PKT)
		{
		    MPIDI_CH3_Pkt_t * pkt = &req->mm.pkt;
		    
		    if (pkt->type < MPIDI_CH3_PKT_END_CH3)
		    {
			MPIDI_DBG_PRINTF((65, FCNAME, "setting mm.send_active"));
			vc->mm.send_active = MPIDI_CH3I_SendQ_head(vc);
		    }
		    else
		    {
			MPIDI_DBG_PRINTF((71, FCNAME, "unknown packet type %d", pkt->type));
		    }
		}
		else if (ca < MPIDI_CH3_CA_END_CH3)
		{
		    MPIDI_DBG_PRINTF((65, FCNAME, "finished sending iovec, calling CH3U_Handle_send_req()"));
		    MPIDI_CH3U_Handle_send_req(vc, req);
		    if (vc->mm.send_active == NULL)
		    {
			/* NOTE: This code assumes that if another write is not posted by the device during the callback, then the
			   device has completed the current request.  As a result, the current request is dequeded and next request
			   in the queue is processed. */
			MPIDI_DBG_PRINTF((65, FCNAME, "request (assumed) complete"));
			MPIDI_DBG_PRINTF((65, FCNAME, "dequeuing req and posting next send"));
			MPIDI_CH3I_SendQ_dequeue(vc);
			vc->mm.send_active = MPIDI_CH3I_SendQ_head(vc);
		    }
		}
		else
		{
		    MPIDI_DBG_PRINTF((65, FCNAME, "ca = %d", ca));
		    assert(ca < MPIDI_CH3I_CA_END_MM_CHANNEL);
		}
	    }
	    else
	    {
		MPIDI_DBG_PRINTF((65, FCNAME, "iovec updated by %d bytes but not complete", nb));
		assert(req->mm.iov_offset < req->ch3.iov_count);
		break;
	    }
	}
	else
	{
	    MPIDI_DBG_PRINTF((65, FCNAME, "shm_post_writev returned %d bytes", nb));
	    break;
	}
    }

    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));

    MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_SHM_WRITTEN);
}

#undef FUNCNAME
#define FUNCNAME handle_sock_op
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int handle_sock_op(sock_event_t *event_ptr)
{
    int rc;
    MPIDI_STATE_DECL(MPID_STATE_HANDLE_SOCK_OP);

    MPIDI_FUNC_ENTER(MPID_STATE_HANDLE_SOCK_OP);
    switch (event_ptr->op_type)
    {
    case SOCK_OP_READ:
	{
	    MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event_ptr->user_ptr;

	    MPID_Request * rreq = conn->recv_active;

	    if (event_ptr->error != SOCK_SUCCESS)
	    {
		connection_recv_fail(conn, event_ptr->error);
		break;
	    }

	    if (conn->recv_active)
	    {
		conn->recv_active = NULL;
		MPIDI_CH3U_Handle_recv_req(conn->vc, rreq);
		if (conn->recv_active == NULL)
		{ 
		    connection_post_recv_pkt(conn);
		}
	    }
	    else /* incoming packet header */
	    {

		if (conn->pkt.type < MPIDI_CH3_PKT_END_CH3)
		{
		    conn->recv_active = NULL;
		    MPIDI_CH3U_Handle_recv_pkt(conn->vc, &conn->pkt);
		    if (conn->recv_active == NULL)
		    { 
			connection_post_recv_pkt(conn);
		    }
		}
		else if (conn->pkt.type == MPIDI_CH3I_PKT_SC_OPEN_REQ)
		{
		    int pg_id;
		    int pg_rank;
		    MPIDI_VC * vc;

		    pg_id = conn->pkt.sc_open_req.pg_id;
		    pg_rank = conn->pkt.sc_open_req.pg_rank;
		    vc = &MPIDI_CH3I_Process.pg->vc_table[pg_rank]; /* FIXME: need to lookup process group from pg_id */
		    assert(vc->mm.pg_rank == pg_rank);

		    if (vc->mm.conn == NULL || MPIR_Process.comm_world->rank < pg_rank)
		    {
			vc->mm.state = MPIDI_CH3I_VC_STATE_CONNECTING;
			vc->mm.sock = conn->sock;
			vc->mm.conn = conn;
			conn->vc = vc;

			conn->pkt.type = MPIDI_CH3I_PKT_SC_OPEN_RESP;
			conn->pkt.sc_open_resp.ack = TRUE;
		    }
		    else
		    {
			conn->pkt.type = MPIDI_CH3I_PKT_SC_OPEN_RESP;
			conn->pkt.sc_open_resp.ack = FALSE;
		    }

		    conn->state = CONN_STATE_OPEN_LSEND;
		    connection_post_send_pkt(conn);

		}
		else if (conn->pkt.type == MPIDI_CH3I_PKT_SC_OPEN_RESP)
		{
		    if (conn->pkt.sc_open_resp.ack)
		    {
			conn->state = CONN_STATE_CONNECTED;
			conn->vc->mm.state = MPIDI_CH3I_VC_STATE_CONNECTED;
			assert(conn->vc->mm.conn == conn);
			assert(conn->vc->mm.sock == conn->sock);

			connection_post_recv_pkt(conn);
			connection_post_sendq_req(conn);
		    }
		    else
		    {
			conn->vc = NULL;
			conn->state = CONN_STATE_CLOSING;
			sock_post_close(conn->sock);
		    }
		}
		else
		{
		    int mpi_errno;

		    mpi_errno = MPIR_Err_create_code(MPI_ERR_INTERN, "[ch3:mm] received packet of unknown type %d",
			conn->pkt.type);
		    MPID_Abort(NULL, mpi_errno);
		}
	    }

	    break;
	}

    case SOCK_OP_WRITE:
	{
	    MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event_ptr->user_ptr;

	    if (event_ptr->error != SOCK_SUCCESS)
	    {
		connection_send_fail(conn, event_ptr->error);
		break;
	    }

	    if (conn->send_active)
	    {
		MPID_Request * sreq = conn->send_active;

		conn->send_active = NULL;
		MPIDI_CH3U_Handle_send_req(conn->vc, sreq);
		if (conn->send_active == NULL)
		{
		    MPIDI_CH3I_SendQ_dequeue(conn->vc);
		    connection_post_sendq_req(conn);
		}
	    }
	    else /* finished writing internal packet header */
	    {
		if (conn->state == CONN_STATE_OPEN_CSEND)
		{
		    /* finished sending open request packet */
		    /* post receive for open response packet */
		    conn->state = CONN_STATE_OPEN_CRECV;
		    connection_post_recv_pkt(conn);
		}
		else if (conn->state == CONN_STATE_OPEN_LSEND)
		{
		    /* finished sending open response packet */
		    if (conn->pkt.sc_open_resp.ack == TRUE)
		    { 
			/* post receive for packet header */
			conn->state = CONN_STATE_CONNECTED;
			conn->vc->mm.state = MPIDI_CH3I_VC_STATE_CONNECTED;
			connection_post_recv_pkt(conn);
			connection_post_sendq_req(conn);
		    }
		    else
		    {
			/* head-to-head connections - close this connection */
			conn->state = CONN_STATE_CLOSING;
			rc = sock_post_close(conn->sock);
			assert(rc == SOCK_SUCCESS);
		    }
		}
	    }

	    break;
	}

    case SOCK_OP_ACCEPT:
	{
	    MPIDI_CH3I_Connection_t * conn;

	    conn = connection_alloc();
	    rc = sock_accept(listener_conn->sock, sock_set, conn, &conn->sock);
	    if (rc == SOCK_SUCCESS)
	    { 
		conn->vc = NULL;
		conn->state = CONN_STATE_OPEN_LRECV;
		conn->send_active = NULL;
		conn->recv_active = NULL;

		connection_post_recv_pkt(conn);
	    }
	    else
	    {
		connection_free(conn);
	    }

	    break;
	}

    case SOCK_OP_CONNECT:
	{
	    MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event_ptr->user_ptr;

	    if (event_ptr->error != SOCK_SUCCESS)
	    {
		int mpi_errno;

		mpi_errno = MPIR_Err_create_code(MPI_ERR_OTHER, "[ch3:mm] failed to connnect to process %d:%d",
		    /* FIXME: pgid */ -1, conn->vc->mm.pg_rank);
		    MPID_Abort(NULL, mpi_errno);

	    }

	    conn->state = CONN_STATE_OPEN_CSEND;
	    conn->pkt.type = MPIDI_CH3I_PKT_SC_OPEN_REQ;
	    conn->pkt.sc_open_req.pg_id = -1; /* FIXME: multiple process groups may exist */
	    conn->pkt.sc_open_req.pg_rank = MPIR_Process.comm_world->rank;
	    connection_post_send_pkt(conn);

	    break;
	}

    case SOCK_OP_CLOSE:
	{
	    MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event_ptr->user_ptr;

	    /* If the conn pointer is NULL then the close was intentional */
	    if (conn != NULL)
	    {
		if (conn->state == CONN_STATE_CLOSING)
		{
		    assert(conn->send_active == NULL);
		    assert(conn->recv_active == NULL);
		    conn->sock = SOCK_INVALID_SOCK;
		    conn->state = CONN_STATE_CLOSED;
		    connection_free(conn);
		}
		else if (conn->state == CONN_STATE_LISTENING && shutting_down)
		{
		    assert(listener_conn == conn);
		    connection_free(listener_conn);
		    listener_conn = NULL;
		    listener_port = 0;
		}
		else
		{ 
		    /* FIXME: an error has occurred */
		}
	    }
	    break;
	}
    }
    MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_SOCK_OP);
    return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Progress(int is_blocking)
{
    int rc;
    int register count;
    sock_event_t event;
    unsigned completions = MPIDI_CH3I_progress_completions;
    MPIDI_CH3I_Shmem_queue_info info;
    int num_bytes;
    MPIDI_VC *vc_ptr;
    int error;
    shm_wait_t shm_op;
    int spin_count = 1;
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS);
    MPIDI_STATE_DECL(MPID_STATE_MPIDU_YIELD);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS);

    MPIDI_DBG_PRINTF((50, FCNAME, "entering, blocking=%s", is_blocking ? "true" : "false"));
    do
    {
	/* make progress on the shared memory queues */

	if (MPIDI_CH3I_Process.shm_reading_list)
	{
	    rc = MPIDI_CH3I_SHM_wait(MPIDI_CH3I_Process.shm_reading_list, 0, &vc_ptr, &num_bytes, &shm_op, &error);
	    assert(rc == MPI_SUCCESS);
	    switch (shm_op)
	    {
	    case SHM_WAIT_TIMEOUT:
		if (spin_count >= MPIDI_CH3I_Process.pg->nShmWaitSpinCount)
		{
#ifdef USE_SLEEP_YIELD
		    if (spin_count >= MPIDI_CH3I_Process.pg->nShmWaitYieldCount)
		    {
			MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_SLEEP_YIELD);
			MPIDU_Sleep_yield();
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_SLEEP_YIELD);
		    }
		    else
		    {
			MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
			MPIDU_Yield();
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
		    }
#else
		    MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
		    MPIDU_Yield();
		    MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
#endif
		}
		spin_count++;
		break;
	    case SHM_WAIT_READ:
		MPIDI_DBG_PRINTF((50, FCNAME, "MPIDI_CH3I_SHM_wait reported %d bytes read", num_bytes));
		spin_count = 1;
#ifdef USE_SLEEP_YIELD
		MPIDI_Sleep_yield_count = 0;
#endif
		handle_shm_read(vc_ptr, num_bytes);
		break;
	    case SHM_WAIT_WRITE:
		MPIDI_DBG_PRINTF((50, FCNAME, "MPIDI_CH3I_SHM_wait reported %d bytes written", num_bytes));
		spin_count = 1;
#ifdef USE_SLEEP_YIELD
		MPIDI_Sleep_yield_count = 0;
#endif
		handle_shm_written(vc_ptr);
		break;
	    case SHM_WAIT_ERROR:
		MPIDI_err_printf("MPIDI_CH3_Progress", "MPIDI_CH3I_SHM_wait returned error %d\n", error);
		break;
	    default:
		assert(FALSE);
		break;
	    }
	}
	else
	{
	    spin_count++;
	}

	if (MPIDI_CH3I_Process.shm_writing_list)
	{
	    /* pound on the write queues since shm_wait currently does not return SHM_WAIT_WRITE */
	    vc_ptr = MPIDI_CH3I_Process.shm_writing_list;
	    while (vc_ptr)
	    {
		if (vc_ptr->mm.send_active != NULL)
		{
		    handle_shm_written(vc_ptr);
		}
		vc_ptr = vc_ptr->mm.shm_next_writer;
	    }
	}

	if (spin_count > (MPIDI_CH3I_Process.pg->nShmWaitSpinCount >> 1) )
	{
	    /* make progress on the sockets */

	    rc = sock_wait(sock_set, 0, &event);
	    if (rc == SOCK_SUCCESS)
	    {
		rc = handle_sock_op(&event);
		assert(rc == MPI_SUCCESS);
	    }
	    else
	    {
		if (rc != SOCK_ERR_TIMEOUT)
		{
		    goto fn_exit;
		}
		MPIDU_Yield();
	    }

	    /* check for new shmem queue connection requests */
	    rc = MPIDI_CH3I_BootstrapQ_recv_msg(MPIDI_CH3I_Process.pg->bootstrapQ, &info, sizeof(info), &num_bytes, FALSE);
	    assert(rc == MPI_SUCCESS);
	    assert(num_bytes == 0 || num_bytes == sizeof(info));
	    if (num_bytes)
	    {
		vc_ptr = &MPIDI_CH3I_Process.pg->vc_table[info.pg_rank];
		rc = MPIDI_CH3I_SHM_Attach_to_mem(&info.info, &vc_ptr->mm.shm_read_queue_info);
		if (rc != MPI_SUCCESS)
		{
		    printf("MPIDI_CH3I_SHM_Attach_to_mem failed, error %d\n", vc_ptr->mm.shm_read_queue_info.error);fflush(stdout);
		}
		assert(rc == MPI_SUCCESS);
		/*printf("shmem accepted\n");fflush(stdout);*/
		/*vc_ptr->mm.state = MPIDI_CH3I_VC_STATE_CONNECTED;*/ /* we are read connected but not write connected */
		vc_ptr->mm.bShm = TRUE;
		vc_ptr->mm.read_shmq = info.info.addr;
		vc_ptr->mm.shm_reading_pkt = TRUE;
		/* add this VC to the global list to be shm_waited on */
		vc_ptr->mm.shm_next_reader = MPIDI_CH3I_Process.shm_reading_list;
		MPIDI_CH3I_Process.shm_reading_list = vc_ptr;
	    }
	}
    }
    while (completions == MPIDI_CH3I_progress_completions && is_blocking);

fn_exit:
    count = MPIDI_CH3I_progress_completions - completions;
    MPIDI_DBG_PRINTF((50, FCNAME, "exiting, count=%d", count));
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS);
    return count;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_poke
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Progress_poke()
{
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
    MPIDI_CH3_Progress(FALSE);
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_end
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Progress_end()
{
    /* MT: This function is empty for the single-threaded implementation */
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_END);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_END);
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_END);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Progress_init()
{
    int rc;
    sock_t sock;
    int mpi_errno = MPI_SUCCESS;
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);

    MPIDI_DBG_PRINTF((60, FCNAME, "entering"));

    rc = sock_init();
    if (rc != SOCK_SUCCESS)
    {
	mpi_errno = sock_errno_to_mpi_errno(rc);
	MPID_Abort(NULL, mpi_errno);
	goto fn_exit;
    }

    /* create sock set */
    rc = sock_create_set(&sock_set);
    if (rc != SOCK_SUCCESS)
    {
	mpi_errno = sock_errno_to_mpi_errno(rc);
	MPID_Abort(NULL, mpi_errno);
	goto fn_exit;
    }

    /* establish non-blocking listener */
    listener_conn = connection_alloc();
    listener_conn->sock = NULL;
    listener_conn->vc = NULL;
    listener_conn->state = CONN_STATE_LISTENING;
    listener_conn->send_active = NULL;
    listener_conn->recv_active = NULL;

    rc = sock_listen(sock_set, listener_conn, &listener_port, &sock);
    if (rc != SOCK_SUCCESS)
    {
	mpi_errno = sock_errno_to_mpi_errno(rc);
	MPID_Abort(NULL, mpi_errno);
	goto fn_exit;
    }

    listener_conn->sock = sock;

fn_exit:
    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);
    return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Progress_finalize()
{
    /* MT: in a multi-threaded environment, finalize() should signal any thread(s) blocking on sock_wait() and wait for those
       threads to complete before destroying the progress engine data structures.  We may need to add a function to the sock
       interface for this. */

    int rc;
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);
    MPIDI_DBG_PRINTF((60, FCNAME, "entering"));

    MPIR_Nest_incr();
    {
	NMPI_Barrier(MPI_COMM_WORLD); /* FIXME: this barrier may not be necessary */
	shutting_down = TRUE;
	NMPI_Barrier(MPI_COMM_WORLD);
    }
    MPIR_Nest_decr();

    /* Shut down the listener */
    rc = sock_post_close(listener_conn->sock);
    assert(rc == SOCK_SUCCESS);

    /* XXX: Wait for listener sock to close */

    /* FIXME: Cleanly shutdown other socks and free connection structures. (close protocol?) */

    sock_destroy_set(sock_set);
    sock_finalize();

    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);
    return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Listener_get_port
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Listener_get_port()
{
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_LISTENER_GET_PORT);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_LISTENER_GET_PORT);
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_LISTENER_GET_PORT);
    return listener_port;
}

static unsigned int GetIP(char *pszIP)
{
    unsigned int nIP;
    unsigned int a,b,c,d;
    if (pszIP == NULL)
	return 0;
    sscanf(pszIP, "%u.%u.%u.%u", &a, &b, &c, &d);
    /*printf("mask: %u.%u.%u.%u\n", a, b, c, d);fflush(stdout);*/
    nIP = (d << 24) | (c << 16) | (b << 8) | a;
    return nIP;
}

static unsigned int GetMask(char *pszMask)
{
    int i, nBits;
    unsigned int nMask = 0;
    unsigned int a,b,c,d;

    if (pszMask == NULL)
	return 0;

    if (strstr(pszMask, "."))
    {
	sscanf(pszMask, "%u.%u.%u.%u", &a, &b, &c, &d);
	/*printf("mask: %u.%u.%u.%u\n", a, b, c, d);fflush(stdout);*/
	nMask = (d << 24) | (c << 16) | (b << 8) | a;
    }
    else
    {
	nBits = atoi(pszMask);
	for (i=0; i<nBits; i++)
	{
	    nMask = nMask << 1;
	    nMask = nMask | 0x1;
	}
    }
    /*
    unsigned int a, b, c, d;
    a = ((unsigned char *)(&nMask))[0];
    b = ((unsigned char *)(&nMask))[1];
    c = ((unsigned char *)(&nMask))[2];
    d = ((unsigned char *)(&nMask))[3];
    printf("mask: %u.%u.%u.%u\n", a, b, c, d);fflush(stdout);
    */
    return nMask;
}

static int GetHostAndPort(char *host, int *port, char *business_card)
{
    char pszNetMask[50];
    char *pEnv, *token;
    unsigned int nNicNet, nNicMask;
    char *temp, *pszHost, *pszIP, *pszPort;
    unsigned int ip;

    pEnv = getenv("MPICH_NETMASK");
    if (pEnv != NULL)
    {
	MPIU_Strncpy(pszNetMask, pEnv, 50);
	token = strtok(pszNetMask, "/");
	if (token != NULL)
	{
	    token = strtok(NULL, "\n");
	    if (token != NULL)
	    {
		nNicNet = GetIP(pszNetMask);
		nNicMask = GetMask(token);

		/* parse each line of the business card and match the ip address with the network mask */
		temp = MPIU_Strdup(business_card);
		token = strtok(temp, ":\r\n");
		while (token)
		{
		    pszHost = token;
		    pszIP = strtok(NULL, ":\r\n");
		    pszPort = strtok(NULL, ":\r\n");
		    ip = GetIP(pszIP);
		    /*msg_printf("masking '%s'\n", pszIP);*/
		    if ((ip & nNicMask) == nNicNet)
		    {
			/* the current ip address matches the requested network so return these values */
			MPIU_Strncpy(host, pszIP, MAXHOSTNAMELEN); /*pszHost);*/
			*port = atoi(pszPort);
			MPIU_Free(temp);
			return MPI_SUCCESS;
		    }
		    token = strtok(NULL, ":\r\n");
		}
		if (temp)
		    MPIU_Free(temp);
	    }
	}
    }

    temp = strdup(business_card);
    if (temp == NULL)
    {
	/*MPIDI_err_printf("GetHostAndPort", "MPIU_Strdup failed\n");*/
	return MPIR_Err_create_code(MPI_ERR_OTHER, "**ch3|sock|strdup", 0);
    }
    /* move to the host part */
    token = strtok(temp, ":");
    if (token == NULL)
    {
	free(temp);
	/*MPIDI_err_printf("GetHostAndPort", "invalid business card\n");*/
	return MPIR_Err_create_code(MPI_ERR_OTHER, "[ch3:mm] GetHostAndPort: Invalid business card - %s", business_card);
    }
    /*strcpy(host, token);*/
    /* move to the ip part */
    token = strtok(NULL, ":");
    if (token == NULL)
    {
	free(temp);
	/*MPIDI_err_printf("GetHostAndPort", "invalid business card\n");*/
	return MPIR_Err_create_code(MPI_ERR_OTHER, "[ch3:mm] GetHostAndPort: Invalid business card - %s", business_card);
    }
    MPIU_Strncpy(host, token, MAXHOSTNAMELEN); /* use the ip string instead of the hostname, it's more reliable */
    /* move to the port part */
    token = strtok(NULL, ":");
    if (token == NULL)
    {
	free(temp);
	/*MPIDI_err_printf("GetHostAndPort", "invalid business card\n");*/
	return MPIR_Err_create_code(MPI_ERR_OTHER, "[ch3:mm] GetHostAndPort: Invalid business card - %s", business_card);
    }
    *port = atoi(token);
    free(temp);

    return MPI_SUCCESS;
}

int MPIDI_CH3I_Shm_connect(MPIDI_VC *vc, char *business_card, int *flag)
{
    int error;
    char hostname[256];
    char *queue_name;
    MPIDI_CH3I_BootstrapQ queue;
    MPIDI_CH3I_Shmem_queue_info shm_info/*, shm_remote_info*/;
    /*int num_bytes;*/
    int i;

    error = MPIU_Strncpy(hostname, business_card, 256);
    queue_name = strtok(hostname, ":");
    if (queue_name == NULL)
    {
	*flag = FALSE;
	return MPI_SUCCESS;
    }
    queue_name = strtok(NULL, "\n");

    if (strcmp(MPIDI_CH3I_Process.pg->shm_hostname, hostname) != 0)
    {
	*flag = FALSE;
	/*printf("%s != %s\n", MPIDI_CH3I_Process.pg->shm_hostname, hostname);*/
	return MPI_SUCCESS;
    }

    *flag = TRUE;
    /*printf("%s == %s\n", MPIDI_CH3I_Process.pg->shm_hostname, hostname);*/

    /*printf("attaching to queue: %s\n", queue_name);fflush(stdout);*/
    error = MPIDI_CH3I_BootstrapQ_attach(queue_name, &queue);
    if (error != MPI_SUCCESS)
    {
	*flag = FALSE;
	return MPI_SUCCESS;
    }
    /*assert(error == MPI_SUCCESS);*/

    /* create the write queue */
    error = MPIDI_CH3I_SHM_Get_mem(sizeof(MPIDI_CH3I_SHM_Queue_t), &vc->mm.shm_write_queue_info);
    assert(error == MPI_SUCCESS);
    vc->mm.write_shmq = vc->mm.shm_write_queue_info.addr;
    vc->mm.write_shmq->head_index = 0;
    vc->mm.write_shmq->tail_index = 0;
    for (i=0; i<MPIDI_CH3I_NUM_PACKETS; i++)
    {
	vc->mm.write_shmq->packet[i].cur_pos = vc->mm.write_shmq->packet[i].data;
	vc->mm.write_shmq->packet[i].avail = MPIDI_CH3I_PKT_AVAILABLE;
    }

    /* send the queue connection information */
    /*printf("copying queue info: size - %d, name - %s\n", vc->mm.shm_write_queue_info.size, vc->mm.shm_write_queue_info.key);fflush(stdout);*/
    shm_info.info = vc->mm.shm_write_queue_info;
    shm_info.pg_id = 0;
    shm_info.pg_rank = MPIR_Process.comm_world->rank;
    /*printf("sending queue info: size - %d, name - %s\n", shm_info.info.size, shm_info.info.key);fflush(stdout);*/
    error = MPIDI_CH3I_BootstrapQ_send_msg(queue, &shm_info, sizeof(shm_info));
    assert(error == MPI_SUCCESS);

    /* free the queue resource */
    /*printf("detaching from queue: %s\n", queue_name);fflush(stdout);*/
    error = MPIDI_CH3I_BootstrapQ_detach(queue);
    assert(error == MPI_SUCCESS);

    /**flag = FALSE;*/
    return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_VC_post_connect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_VC_post_connect(MPIDI_VC * vc)
{
    char * key;
    char * val;
    int key_max_sz;
    int val_max_sz;
    char host[MAXHOSTNAMELEN];
    int port;
    int rc;
    MPIDI_CH3I_Connection_t * conn;
    int mpi_errno = MPI_SUCCESS;
    int connected;
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);

    MPIDI_DBG_PRINTF((60, FCNAME, "entering"));

    assert(vc->mm.state == MPIDI_CH3I_VC_STATE_UNCONNECTED);

    vc->mm.state = MPIDI_CH3I_VC_STATE_CONNECTING;

    key_max_sz = PMI_KVS_Get_key_length_max();
    key = MPIU_Malloc(key_max_sz);
    assert(key != NULL);
    val_max_sz = PMI_KVS_Get_value_length_max();
    val = MPIU_Malloc(val_max_sz);
    assert(val != NULL);

    rc = snprintf(key, key_max_sz, "P%d-shm_businesscard", vc->mm.pg_rank);
    assert(rc > -1 && rc < key_max_sz);
    rc = PMI_KVS_Get(vc->mm.pg->kvs_name, key, val);
    assert(rc == MPI_SUCCESS);

    connected = FALSE;
    rc = MPIDI_CH3I_Shm_connect(vc, val, &connected);
    assert(rc == MPI_SUCCESS);
    if (connected)
    {
	MPIU_Free(val);
	MPIU_Free(key);

	/*printf("shmem connected\n");fflush(stdout);*/
	vc->mm.shm_next_writer = MPIDI_CH3I_Process.shm_writing_list;
	MPIDI_CH3I_Process.shm_writing_list = vc;

	vc->mm.state = MPIDI_CH3I_VC_STATE_CONNECTED;
	vc->mm.shm_reading_pkt = TRUE;
	vc->mm.send_active = MPIDI_CH3I_SendQ_head(vc); /* MT */

	MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);
	return mpi_errno;
    }

    rc = snprintf(key, key_max_sz, "P%d-sock_businesscard", vc->mm.pg_rank);
    assert(rc > -1 && rc < key_max_sz);
    rc = PMI_KVS_Get(vc->mm.pg->kvs_name, key, val);
    assert(rc == MPI_SUCCESS);

    rc = GetHostAndPort(host, &port, val);
    assert(rc == MPI_SUCCESS);
    /*printf("Connecting to: host %s, port %d\n", host, port);fflush(stdout);*/

    conn = connection_alloc();

    /*rc = sock_post_connect(sock_set, conn, val, port, &conn->sock);*/
    rc = sock_post_connect(sock_set, conn, host, port, &conn->sock);
    if (rc == SOCK_SUCCESS)
    {
	vc->mm.sock = conn->sock;
	vc->mm.conn = conn;
	conn->vc = vc;
	conn->state = CONN_STATE_CONNECTING;
	conn->send_active = NULL;
	conn->recv_active = NULL;
    }
    else
    {
	if (rc == SOCK_ERR_HOST_LOOKUP)
	{ 
	    mpi_errno = MPIR_Err_create_code(MPI_ERR_OTHER, "[ch3:mm] failed to obtain host information for process %d:%d (%s)",
					     /* FIXME: pgid*/ -1, conn->vc->mm.pg_rank, val);
	}
	else if (rc == SOCK_ERR_CONN_REFUSED)
	{ 
	    mpi_errno = MPIR_Err_create_code(MPI_ERR_OTHER, "[ch3:mm] failed to connect to process %d:%d (%s)",
					     /* FIXME: pgid */ -1, conn->vc->mm.pg_rank, val);
	}
	else
	{
	    mpi_errno = sock_errno_to_mpi_errno(rc);
	}

	MPID_Abort(NULL, mpi_errno);

	vc->mm.state = MPIDI_CH3I_VC_STATE_FAILED;
	connection_free(conn);
    }

    MPIU_Free(val);
    MPIU_Free(key);

    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);
    return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_MM_VC_post_read
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3I_MM_VC_post_read(MPIDI_VC * vc, MPID_Request * rreq)
{
    MPIDI_CH3I_Connection_t * conn = vc->mm.conn;
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_VC_POST_READ);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_VC_POST_READ);
    MPIDI_DBG_PRINTF((60, FCNAME, "entering, vc=0x%08x, rreq=0x%08x", (unsigned) vc, rreq->handle));

    assert (conn->recv_active == NULL);
    conn->recv_active = rreq;
    sock_post_readv(conn->sock, rreq->ch3.iov + rreq->mm.iov_offset, rreq->ch3.iov_count - rreq->mm.iov_offset, NULL);

    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_READ);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_MM_VC_post_write
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3I_MM_VC_post_write(MPIDI_VC * vc, MPID_Request * sreq)
{
    MPIDI_CH3I_Connection_t * conn = vc->mm.conn;
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_VC_POST_WRITE);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_VC_POST_WRITE);
    MPIDI_DBG_PRINTF((60, FCNAME, "entering, vc=0x%08x, sreq=0x%08x", (unsigned) vc, sreq->handle));

    assert (conn->send_active == NULL);
    conn->send_active = sreq;
    if (vc->mm.bShm)
    {
	printf("help me i'm melting.\n");
    }
    else
    {
	sock_post_writev(conn->sock, sreq->ch3.iov + sreq->mm.iov_offset, sreq->ch3.iov_count - sreq->mm.iov_offset, NULL);
    }

    MPIDI_DBG_PRINTF((60, FCNAME, "exiting"));
    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_WRITE);
}

#undef FUNCNAME
#define FUNCNAME connection_alloc
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline MPIDI_CH3I_Connection_t * connection_alloc(void)
{
    MPIDI_CH3I_Connection_t * conn;
    int mpi_errno;

    MPIDI_STATE_DECL(MPID_STATE_CONNECTION_ALLOC);

    MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_ALLOC);
    conn = MPIU_Malloc(sizeof(MPIDI_CH3I_Connection_t));
    if (conn == NULL)
    {
	mpi_errno = MPIR_Err_create_code( MPI_ERR_OTHER, "ch3:mm unable to allocate a connection structure", 0 );
	MPID_Abort(NULL, mpi_errno);
    }

    MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_ALLOC);
    return conn;
}

#undef FUNCNAME
#define FUNCNAME connection_free
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void connection_free(MPIDI_CH3I_Connection_t * conn)
{
    MPIDI_STATE_DECL(MPID_STATE_CONNECTION_FREE);

    MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_FREE);

    MPIU_Free(conn);

    MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_FREE);
}

#undef FUNCNAME
#define FUNCNAME connection_post_sendq_req
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void connection_post_sendq_req(MPIDI_CH3I_Connection_t * conn)
{
    MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SENDQ_REQ);

    MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SENDQ_REQ);
    /* post send of next request on the send queue */
    conn->send_active = MPIDI_CH3I_SendQ_head(conn->vc); /* MT */
    if (conn->send_active != NULL)
    {
	int rc;

	rc = sock_post_writev(conn->sock, conn->send_active->ch3.iov, conn->send_active->ch3.iov_count, NULL);
	if (rc != SOCK_SUCCESS)
	{
	    connection_send_fail(conn, rc);
	}
    }

    MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SENDQ_REQ);
}

#undef FUNCNAME
#define FUNCNAME connection_post_send_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void connection_post_send_pkt(MPIDI_CH3I_Connection_t * conn)
{
    int rc;
    MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SEND_PKT);

    MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SEND_PKT);

    rc = sock_post_write(conn->sock, &conn->pkt, sizeof(conn->pkt), NULL);
    if (rc != SOCK_SUCCESS)
    {
	connection_send_fail(conn, rc);
    }

    MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SEND_PKT);
}

#undef FUNCNAME
#define FUNCNAME connection_post_recv_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void connection_post_recv_pkt(MPIDI_CH3I_Connection_t * conn)
{
    int rc;
    MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_RECV_PKT);

    MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_RECV_PKT);

    rc = sock_post_read(conn->sock, &conn->pkt, sizeof(conn->pkt), NULL);
    if (rc != SOCK_SUCCESS)
    {
	connection_recv_fail(conn, rc);
    }

    MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_RECV_PKT);
}

#undef FUNCNAME
#define FUNCNAME connection_send_fail
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void connection_send_fail(MPIDI_CH3I_Connection_t * conn, int sock_errno)
{
    int mpi_errno;
    MPIDI_STATE_DECL(MPID_STATE_CONNECTION_SEND_FAIL);

    MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_SEND_FAIL);

    mpi_errno = sock_errno_to_mpi_errno(sock_errno);

#   if 0
    {
	conn->state = CONN_STATE_FAILED;
	if (conn->vc != NULL)
	{
	    
	    conn->vc->mm.state = MPIDI_CH3I_VC_STATE_FAILED;
	    MPIDI_CH3U_VC_send_failure(conn->vc, mpi_errno);
	}
    }
#   endif

    if (conn->send_active)
    { 
	MPID_Abort(conn->send_active->comm, mpi_errno);
    }
    else
    {
	MPID_Abort(NULL, mpi_errno);
    }

    MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_SEND_FAIL);
}

#undef FUNCNAME
#define FUNCNAME connection_recv_fail
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void connection_recv_fail(MPIDI_CH3I_Connection_t * conn, int sock_errno)
{
    int mpi_errno;
    MPIDI_STATE_DECL(MPID_STATE_CONNECTION_RECV_FAIL);

    MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_RECV_FAIL);

    mpi_errno = sock_errno_to_mpi_errno(sock_errno);
    MPID_Abort(NULL, mpi_errno);

    MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_RECV_FAIL);
}

static int sock_errno_to_mpi_errno(int sock_errno)
{
    int mpi_errno;

    switch(sock_errno)
    {
	case SOCK_EOF:
	    mpi_errno = MPIR_Err_create_code( MPI_ERR_OTHER, "[ch3:mm] active connection unexpectedly closed", 0 );
	    break;
	case SOCK_ERR_NOMEM:
	    mpi_errno = MPIR_Err_create_code( MPI_ERR_OTHER, "[ch3:mm] memory allocation failure", 0 );
	    break;
	case SOCK_ERR_HOST_LOOKUP:
	    mpi_errno = MPIR_Err_create_code( MPI_ERR_OTHER, "[ch3:mm] hostname lookup failed", 0 );
	    break;
	case SOCK_ERR_CONN_REFUSED:
	    mpi_errno = MPIR_Err_create_code( MPI_ERR_OTHER, "[ch3:mm] connection refused", 0 );
	    break;
	case SOCK_ERR_CONN_FAILED:
	    mpi_errno = MPIR_Err_create_code( MPI_ERR_OTHER, "[ch3:mm] active connection unexpectedly terminated", 0 );
	    break;
	case SOCK_ERR_BAD_SOCK:
	    mpi_errno = MPIR_Err_create_code( MPI_ERR_INTERN, "[ch3:mm] internal error - bad sock", 0 );
	    break;
	case SOCK_ERR_BAD_BUFFER:
	    mpi_errno = MPIR_Err_create_code( MPI_ERR_BUFFER, "[ch3:mm] bad user buffer", 0 );
	    break;
	default:
	    mpi_errno = MPIR_Err_create_code( MPI_ERR_OTHER, "[ch3:mm] socket error %d", sock_errno, 0);
	    break;
    }

    return mpi_errno;
}
